﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.DataPreprocessing {
  [Item("StatisticsLogic", "")]
  [StorableClass]
  public class StatisticsLogic : Item {
    [Storable]
    private readonly IPreprocessingData preprocessingData;

    #region Constructor, Cloning & Persistence
    public StatisticsLogic(IPreprocessingData preprocessingData) {
      this.preprocessingData = preprocessingData;
    }
    public StatisticsLogic(StatisticsLogic original, Cloner cloner)
      : base(original, cloner) {
      preprocessingData = cloner.Clone(original.preprocessingData);
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new StatisticsLogic(this, cloner);
    }

    [StorableConstructor]
    protected StatisticsLogic(bool deserializing)
      : base(deserializing) { }
    #endregion

    public int GetMissingValueCount() {
      int count = 0;
      for (int i = 0; i < preprocessingData.Columns; ++i) {
        count += GetMissingValueCount(i);
      }
      return count;
    }

    public int GetMissingValueCount(int columnIndex) {
      return Enumerable.Range(0, preprocessingData.Rows).Count(rowIndex => preprocessingData.IsCellEmpty(columnIndex, rowIndex));
    }

    public int GetRowMissingValueCount(int rowIndex) {
      var columnIndexes = Enumerable.Range(0, preprocessingData.Columns);
      return columnIndexes.Count(columnIndex => !preprocessingData.IsCellEmpty(columnIndex, rowIndex));
    }

    private IEnumerable<T> GetValuesWithoutNaN<T>(int columnIndex, bool considerSelection = false) {
      return preprocessingData.GetValues<T>(columnIndex, considerSelection).Where(x => !PreprocessingData.IsMissingValue(x)).ToList();
    }

    public T GetMin<T>(int columnIndex, T defaultValue, bool considerSelection = false) where T : IComparable<T> {
      var min = defaultValue;
      if (preprocessingData.VariableHasType<T>(columnIndex)) {
        var values = GetValuesWithoutNaN<T>(columnIndex, considerSelection);
        if (values.Any()) {
          min = values.Min();
        }
      }
      return min;
    }

    public T GetMax<T>(int columnIndex, T defaultValue, bool considerSelection = false) where T : IComparable<T> {
      var max = defaultValue;
      if (preprocessingData.VariableHasType<T>(columnIndex)) {
        var values = GetValuesWithoutNaN<T>(columnIndex, considerSelection);
        if (values.Any()) {
          max = values.Max();
        }
      }
      return max;
    }

    public double GetMedian(int columnIndex, bool considerSelection = false) {
      double median = double.NaN;
      if (preprocessingData.VariableHasType<double>(columnIndex)) {
        var values = GetValuesWithoutNaN<double>(columnIndex, considerSelection);
        if (values.Any()) {
          median = values.Median();
        }
      }
      return median;
    }

    public double GetAverage(int columnIndex, bool considerSelection = false) {
      double avg = double.NaN;
      if (preprocessingData.VariableHasType<double>(columnIndex)) {
        var values = GetValuesWithoutNaN<double>(columnIndex, considerSelection);
        if (values.Any()) {
          avg = values.Average();
        }
      }
      return avg;
    }

    public DateTime GetMedianDateTime(int columnIndex, bool considerSelection = false) {
      DateTime median = new DateTime();
      if (preprocessingData.VariableHasType<DateTime>(columnIndex)) {
        median = GetSecondsAsDateTime(GetDateTimeAsSeconds(columnIndex, considerSelection).Median());
      }
      return median;
    }

    public DateTime GetAverageDateTime(int columnIndex, bool considerSelection = false) {
      DateTime avg = new DateTime();
      if (preprocessingData.VariableHasType<DateTime>(columnIndex)) {
        avg = GetSecondsAsDateTime(GetDateTimeAsSeconds(columnIndex, considerSelection).Average());
      }
      return avg;
    }

    public T GetMostCommonValue<T>(int columnIndex, T defaultValue, bool considerSelection = false) {
      var values = GetValuesWithoutNaN<T>(columnIndex, considerSelection);
      if (!values.Any())
        return defaultValue;
      return values.GroupBy(x => x)
                              .OrderByDescending(g => g.Count())
                              .Select(g => g.Key)
                              .First();
    }


    public double GetStandardDeviation(int columnIndex) {
      double stdDev = double.NaN;
      if (preprocessingData.VariableHasType<double>(columnIndex)) {
        stdDev = GetValuesWithoutNaN<double>(columnIndex).StandardDeviation();
      } else if (preprocessingData.VariableHasType<DateTime>(columnIndex)) {
        stdDev = GetDateTimeAsSeconds(columnIndex).StandardDeviation();
      }
      return stdDev;
    }

    public double GetVariance(int columnIndex) {
      double variance = double.NaN;
      if (preprocessingData.VariableHasType<double>(columnIndex)) {
        variance = GetValuesWithoutNaN<double>(columnIndex).Variance();
      } else if (preprocessingData.VariableHasType<DateTime>(columnIndex)) {
        variance = GetDateTimeAsSeconds(columnIndex).Variance();
      }
      return variance;
    }

    public double GetOneQuarterPercentile(int columnIndex) {
      double percentile = double.NaN;
      if (preprocessingData.VariableHasType<double>(columnIndex)) {
        percentile = GetValuesWithoutNaN<double>(columnIndex).DefaultIfEmpty(double.NaN).Quantile(0.25);
      } else if (preprocessingData.VariableHasType<DateTime>(columnIndex)) {
        percentile = GetDateTimeAsSeconds(columnIndex).DefaultIfEmpty(double.NaN).Quantile(0.25);
      }
      return percentile;
    }

    public double GetThreeQuarterPercentile(int columnIndex) {
      double percentile = double.NaN;
      if (preprocessingData.VariableHasType<double>(columnIndex)) {
        percentile = GetValuesWithoutNaN<double>(columnIndex).DefaultIfEmpty(double.NaN).Quantile(0.75);
      } else if (preprocessingData.VariableHasType<DateTime>(columnIndex)) {
        percentile = GetDateTimeAsSeconds(columnIndex).DefaultIfEmpty(double.NaN).Quantile(0.75);
      }
      return percentile;
    }

    public int GetDifferentValuesCount<T>(int columnIndex) {
      return preprocessingData.GetValues<T>(columnIndex).GroupBy(x => x).Count();
    }



    public bool VariableHasType<T>(int columnIndex) {
      return preprocessingData.VariableHasType<T>(columnIndex);
    }

    public string GetColumnTypeAsString(int columnIndex) {
      if (preprocessingData.VariableHasType<double>(columnIndex)) {
        return "double";
      } else if (preprocessingData.VariableHasType<string>(columnIndex)) {
        return "string";
      } else if (preprocessingData.VariableHasType<DateTime>(columnIndex)) {
        return "DateTime";
      }
      return "Unknown Type";
    }

    private IEnumerable<double> GetDateTimeAsSeconds(int columnIndex, bool considerSelection = false) {
      return GetValuesWithoutNaN<DateTime>(columnIndex, considerSelection).Select(x => (double)x.Ticks / TimeSpan.TicksPerSecond);
    }


    private DateTime GetSecondsAsDateTime(double seconds) {
      DateTime dateTime = new DateTime();
      return dateTime.AddSeconds(seconds);
    }

    public event DataPreprocessingChangedEventHandler Changed {
      add { preprocessingData.Changed += value; }
      remove { preprocessingData.Changed -= value; }
    }
  }
}
